#!/usr/bin/env ruby
#    Rubyripper - A secure ripper for Linux/BSD/OSX
#    Copyright (C) 2007  Bouke Woudstra (rubyripperdev@gmail.com)
#
#    This file is part of Rubyripper. Rubyripper is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>

require 'optparse'
require 'ostruct'

class Gui_CLI
	

	if File.exists?('rr_lib.rb')
		require './rr_lib.rb' # run from local directory
	else
		require 'rr_lib.rb' # installed in ruby directory
	end
	
	def initialize()
		parse_options()
		@ripping_log = ""
		@ripping_progress = 0.0
		@encoding_progress = 0.0
		@settings = {"flac" => false, "flacsettings" => "--best", "vorbis" => true, "vorbissettings" => "-q 4", "mp3" => false, "mp3settings" => "-V 3", "wav" => false, "other" => false, "othersettings" => '', "playlist" => true, "cdrom" => cdrom_drive(), "offset" => 0, "threading" => false, "rippersettings" => '-Z', "max_tries" => 0, 'basedir' => '~/', 'naming_normal' => '%f/%a (%y) %b/%n - %t', 'naming_various' => '%f/%a (%y) %b/%n - %va - %t', 'verbose' => false, "debug" => false, "instance" => self, "eject" => true, "req_matches_errors" => 2, "req_matches_all" => 2, "site" => "freedb.org", "username" => "anonymous", "hostname" => "my_secret.com", "first_hit" => true, "freedb" => true, "edit" => false, "debug" => true, "verbose" => true} # some sane default settings freedb and EXCEPT THE LAST TWO!!!!
		if File.exist?(File.expand_path(@options.file))
			load_configfile()
			if @options.configure: edit_settings() end
		else
			edit_settings() # Always set up settings if no config file exists.
		end		
		main()
	end

	def parse_options
		@options = OpenStruct.new
		@options.file = "~/.rubyripper/settings"
		@options.help = false
		@options.verbose = false
		@options.configure = false
		@options.all = false
		@options.help = false

		opts = OptionParser.new(banner = nil, width = 15, indent = ' ' * 2) do |opts|
			opts.on("-f", "--file FILE", "Load configuration settings from file <FILE>.") do |f|
				@options.file = f
			end
			opts.on("-v", "--verbose", "Display verbose output.") do |v|
				@options.verbose = v
			end
			opts.on("-c", "--configure", "Change configuration settings.") do |c|
				@options.configure = c
			end
			opts.on("-a", "--all", "Rip all tracks. Skip any questions about track selection.") do |a|
				@options.all = a
			end
			opts.on_tail("-h", "--help", "Show this usage statement") do |h|
				puts opts
				@options.help = h
			end
		end

		begin
			opts.parse!(ARGV)
			rescue Exception => e
			puts e, "", opts
			exit
		end

		if @options.help: exit end

		puts "Verbose output specified." if @options.verbose
		puts "Configure option specified." if @options.configure
		puts "Rip all tracks specified." if @options.all
		puts "Use config file " + @options.file if @options.file
	end
	
	def show_settings # simple overview of current settings, with numbers so a user can choose what to change
		puts "#### RUBYRIPPER SETTINGS ####"
		puts ""
		puts "  ## ENCODING ##  "
 		puts "1)  flac [#{@settings['flac'] ? '*' : ' '}] , settings: #{@settings['flacsettings']}"
 		puts "2)  vorbis [#{@settings['vorbis'] ? '*' : ' '}] , settings: #{@settings['vorbissettings']}"
 		puts "3)  mp3 [#{@settings['mp3'] ? '*' : ' '}] , settings: #{@settings['mp3settings']}"
 		puts "4)  wav [#{@settings['wav'] ? '*' : ' '}]"
 		puts "5)  other [#{@settings['other'] ? '*' : ' '}] , settings: #{@settings['othersettings']}"
 		puts "6)  playlist support [#{@settings['playlist'] ? '*' : ' '}]"
 		puts "7)  encode while ripping [#{@settings['threading'] ? '*' : ' '}]"
 		puts ""
 		puts "  ## RIPPING ##  "
 		puts "8)  cdrom : #{@settings['cdrom']} with offset #{@settings['offset']}"
 		puts "9)  passing extra cdparanoia parameters : #{@settings['rippersettings'] != '' ? @settings['rippersettings'] : 'none'}"
		puts "10) match all chunks : #{@settings['req_matches_all']}"
		puts "11) match erronous chunks : #{@settings['req_matches_errors']}"
		puts "12) maximum trials : #{@settings['max_tries'] == 0 ? "no maximum" : @settings['max_tries']}"
		puts "13) eject disc after ripping [#{@settings['eject'] ? '*' : ' '}]"
		puts ""
		puts " ##NAMING SCHEME ## "
		puts "14) base directory : #{@settings['basedir']}"
		puts "15) standard scheme: #{@settings['naming_normal']}"
		puts "16) various artist scheme: #{@settings['naming_various']}"
 		puts ""
 		puts "  ## FREEDB ##  "
		puts "17) fetch cd info with freedb [#{@settings['freedb'] ? '*' : ' '}]"
		puts "18) always use first hit [#{@settings['first_hit'] ? '*' : ' '}]"
		puts "19) site = #{@settings['site']}"
		puts "20) username = #{@settings['username']}"
		puts "21) hostname = #{@settings['hostname']}"
		puts "22) edit freedb data before ripping [#{@settings['edit'] ? '*' : ' '}]"
 		puts ""
		puts "88) config file = #{@options.file}"
 		puts "99) Don't change any setting"
 		puts ""
	end

	def edit_settings # depending on the input of the user as asked in show_settings(), change the setting
		show_settings()
		while (number = get_answer("Please type the number of the setting you wish to change : ", "number", 99)) != 99
			if number == 1 : set_codec('flac')
			elsif number == 2 : set_codec('vorbis')
			elsif number == 3 : set_codec('mp3')
			elsif number == 4 : if @settings['wav'] : puts "wav disabled" ; @settings['wav'] = false else puts "wav enabled" ; @settings['wav'] = true end
			elsif number == 5 : set_codec('other')
			elsif number == 6 : if @settings['playlist'] : puts "playlist disabled" ; @settings['playlist'] = false else puts "playlist enabled" ; @settings['playlist'] = true end
			elsif number == 7 : if @settings['threading'] : puts "threading disabled" ; @settings['threading'] = false else puts "threading enabled" ; @settings['threading'] = true end
			elsif number == 8 : set_cdrom()
			elsif number == 9 : @settings['rippersettings'] = get_answer("Which options to pass to cdparanoia? : ", "open", "")
			elsif number == 10 : @settings['req_matches_all'] = get_answer("How many times must all chunks be matched? : ", "number", 2)
			elsif number == 11 : @settings['req_matches_errors'] = get_answer("How many times must erronous chunks be matched? : ", "number", 3)
			elsif number == 12 : @settings['max_tries'] = get_answer("What should be the maximum number of tries? : ", "number", 5)
			elsif number == 13 : if @settings['eject'] : puts "eject disabled" ; @settings['eject'] = false else puts "eject enabled" ; @settings['eject'] = true end
			elsif number == 14 : @settings['basedir'] = get_answer("Please enter your directory for your encodings: ", "open", "")
			elsif number == 15 : setdir('normal')
			elsif number == 16 : setdir('various')
			elsif number == 17 : if @settings['freedb'] : puts "freedb disabled" ; @settings['freedb'] = false else puts "freedb enabled" ; @settings['freedb'] = true end
			elsif number == 18 : if @settings['first_hit'] : puts "first hit disabled" ; @settings['first_hit'] = false else puts "first hit enabled" ; @settings['first_hit'] = true end
			elsif number == 19 : @settings['site'] = get_answer("Freedb mirror: ", "open", "freedb.org")
			elsif number == 20 : @settings['username'] = get_answer("Username : ", "open", "anonymous")
			elsif number == 21 : @settings['hostname'] = get_answer("Hostname : ", "open", "my_secret.com")
			elsif number == 22 : if @settings['edit'] : puts "freedb editing disabled" ; @settings['edit'] = false else puts "freedb editing enabled" ; @settings['edit'] = true end 
			elsif number == 88 : @options.file = get_answer("Config file : ", "open", @options.file)
			else puts "Number #{number} is not an option!\nPlease try again."
			end
			puts ""
			show_settings()
		end
		save_configfile()
		unless get_answer("Do you want to continue with the program now? (y/n) : ", "yes", "y") : exit() end
	end
	
	def set_codec(codec) # little help function for edit_settings
		@settings[codec] = get_answer("Do you want to enable #{codec} encoding? (y/n) : ", "yes", "y")
		if @settings[codec]
			if get_answer("Do you want to change the encoding parameters for #{codec}? (y/n) : ", "yes", "n")
				if codec == 'other' : puts "%a = artist, %b = album, %g = genre, %y = year, %t = trackname, %n = tracknumber, %i = inputfile, %o = outputfile (don't forget extension)" end
				@settings[codec + 'settings'] = get_answer("Encoding paramaters for #{codec} : ", "open", "")
			end
		end
	end
	
	def set_cdrom # little help function for edit_settings
		@settings['cdrom'] = get_answer("Cdrom device : ", "open", cdrom_drive())
		@settings['offset'] = get_answer("Offset for drive : ", "number", 0)
	end

	def setdir(variable)
		if variable == 'normal'
			puts "\nCurrent naming scheme: #{@settings['naming_normal']}"
			puts get_example_filename_normal(@settings['basedir'], @settings['naming_normal'])
		else 
			puts "\nCurrent naming scheme: #{@settings['naming_various']}"
			puts get_example_filename_various(@settings['basedir'], @settings['naming_various'])
		end

		puts	"\n%a = Artist	\n%b = Album\n%g = Genre\n%y = Year\n%f = Codec\n%n = Tracknumber\n%t = Trackname\n%va = Various Artist\n"
		answer = get_answer("New #{variable} naming scheme (q to quit) : ", "open", "%f/%a (%y) %b/%n - %t")
		if answer != 'q'
			if variable == 'normal' : puts "An example filename is now:\n\n\t#{get_example_filename_normal(@settings['basedir'], answer)}" ; @settings['naming_normal'] = answer
			else puts "An example filename is now:\n\n\t#{get_example_filename_various(@settings['basedir'], answer)}" ; @settings['naming_various'] = answer
			end
		end
	end

	def load_configfile # load config file
		if File.exists?(File.expand_path(@options.file))
			file = File.new(File.expand_path(@options.file),'r')
			while line = file.gets
				key, value = line.split('=', 2)
				value.rstrip! # remove the trailing newline character
				if value == "false" : value = false # replace the string with a bool
				elsif value == "true" : value = true  # replace the string with a bool
				elsif value == "''" : value = '' #replace a string that contains two quotes with an empty string
				elsif value[0,1] == "#" : next #name of previous instances should not be loaded
				elsif value.to_i > 0 || value == '0' : value = value.to_i
				end
				@settings[key] = value
			end
			file.close()
		end
	end
	
	def save_configfile # save config file
		if not File.directory?(dirname = File.join(ENV['HOME'], '.rubyripper')) : Dir.mkdir(dirname) end
		file = File.new(File.expand_path(@options.file), 'w')
		@settings.each do |key, value|
			file.puts "#{key}=#{value}"
		end
		file.close()
	end
	
	def get_answer(question, answer, default)
		succes = false
		while !succes 
			if answer == "yes"
				answers = ["yes", "y", "no", "n"]
				STDOUT.print(question + " [#{default}] ")
				input = STDIN.gets.strip
				if input == '' : input = default end
				if answers.include?(input)
					if input == 'y' || input == "yes" : return true else return false end
				else puts "Please answer yes or no" end
			elsif answer == "open"
				STDOUT.print(question + " [#{default}] ")
				input = STDIN.gets.strip
				if input == '' : return default
				else return input end
			elsif answer == "number"
				STDOUT.print(question + " [#{default}] ")
				input = STDIN.gets.strip
				if input == '' : return default end
				input = input.to_i
				if input.size != 0 : return input
				else puts "#{input} is not a valid number! Please enter a valid number" end
			else puts "We should never get here" ; puts "answer = #{answer}, question = #{question} (error)"
 			end
		end
	end

	def main # after the settings are loaded this function is launched
		get_cd_info()
		tracklist() # Which tracks should be ripped?
		@rubyripper = Rubyripper.new(@settings) # starts some check if the settings are sane
		if @rubyripper.settings_ok : @rubyripper.start_rip end
	end

	def get_cd_info
		@settings['cd'] = Cddb.new(@settings['cdrom'], @settings['instance'], @settings['verbose']) # Analyze the TOC of disc in drive
		if @settings['cd'].audiotracks != 0 # a disc is found
			puts "Audio-disc found, number of tracks : #{@settings['cd'].audiotracks}, total playlength : #{@settings['cd'].cd_playtime}"
			if @settings['freedb'] : puts "Fetching freedb info..." ; @settings['cd'].freedb(@settings, !@settings['first_hit']) end # fetch freedb info, multiple hits? -> choose_freedb()
			if @settings['edit'] : edit_freedb() else show_freedb() end
		else 
			puts "No audio-disc found !";
			edit_settings()
		end
	end
	
	def choose_freedb(choices) #callback function initiated by rr_lib.rb via our update function
		puts "Freedb reported multiple possibilities."
		choices.each_index{|index| puts "#{index + 1}) #{choices[index]}"}
		choice = get_answer("Please type the number of the one you prefer? : ", "number", 1)
		@settings['cd'].freedb_choice(choice - 1)
	end

	def show_freedb(edit = false) # leave the letters and numbers away if edit mode == false
		puts ""
		puts "FREEDB INFO"
		puts "#{"a) " if edit}Artist : #{@settings['cd'].artist}"
		puts "#{"b) " if edit}Album : #{@settings['cd'].album}"
		puts "#{"c) " if edit}Genre : #{@settings['cd'].genre}"
		puts "#{"d) " if edit}Year : #{@settings['cd'].year}"
		puts ""
		@settings['cd'].audiotracks.times{|index| puts "#{"#{index + 1}) " if edit}Track #{index + 1} : #{@settings['cd'].tracklist[index]}"}
		puts ""
	end
	
	def edit_freedb
		show_freedb(edit = true)
		puts "99) Don't change any setting"
 		puts ""
		while (answer = get_answer("Please type the letter or number of the setting you wish to change :", "open", "99")) != "99"
			if answer == "a" : @settings['cd'].artist = get_answer("Artist : ", "open", @settings['cd'].artist)
			elsif answer == "b" : @settings['cd'].album = get_answer("Album : ", "open", @settings['cd'].album)
			elsif answer == "c" : @settings['cd'].genre = get_answer("Genre : ", "open", @settings['cd'].genre)
			elsif answer == "d" : @settings['cd'].year = get_answer("Year : ", "open", @settings['cd'].year)
			else
				if answer.to_i > 0 && answer.to_i <= @settings['cd'].audiotracks #Check if tracknumber is in valid range
					@settings['cd'].tracklist[answer.to_i - 1] = get_answer("Track #{answer.to_i} : ", "open", @settings['cd'].tracklist[answer.to_i - 1])
				else
					puts "This is not a valid number. Try again"
				end
			end	
			show_freedb(edit = true)
		end
	end

	def dir_exists
		puts "The output directory already exists. What would you like to do?"
		puts ""
		puts "1) Auto rename the output directory"
		puts "2) Overwrite the existing directory"
		puts "3) Cancel the rip"
		puts ""
		answer = get_answer("Please enter the number of your choice: ", "number", "1")
		if answer == 1: @rubyripper.postfix_dir() ; @rubyripper.start_rip()
		elsif answer == 2: @rubyripper.overwrite_dir() ; @rubyripper.start_rip()
		else puts "Rip is canceled, exiting..." ; exit()
		end
	end

  	def update(modus, value=false)
		if modus == "ripping_progress"
			progress = "%.3g" % (value * 100)
			puts "Ripping progress (#{progress} %)"
		elsif modus == "encoding_progress"
			progress = "%.3g" % (value * 100)
			puts "Encoding progress (#{progress} %)"
		elsif modus == "log_change"
			print value
		elsif modus == "error"
			print value
			print "\n"
			if get_answer("Do you want to change your settings? (y/n) : ", "yes", "y") : edit_settings() end
		elsif modus == "cddb_hit" && value != false
			choose_freedb(value)
		elsif modus == "dir_exists"
			dir_exists()
		end
	end

	def tracklist # Fill @settings['tracks_to_rip']
		@settings['tracks_to_rip'] = Array.new
		@settings['cd'].audiotracks.times{|number| @settings['tracks_to_rip'] << number + 1} # Start with all tracks selected
		if @options.all || get_answer("\nShould all tracks be ripped ? (y/n) ", "yes", 'y')
			puts "Tracks to rip are #{@settings['tracks_to_rip'].join(" ")}"
		else
			succes = false
			while !succes
				puts "Current selection of tracks : #{@settings['tracks_to_rip'].join(' ')}"
				number = get_answer("Enter 1 for entering the tracknumbers you want to remove.\nEnter 2 for entering the tracks you want to keep.\nYour choice: ", "number", 1)
				if number == 1
					print "Type the numbers of the tracks you want to remove and separate them with a space: "
					answer = STDIN.gets.strip.split
					answer.each_index{|index| answer[index] = answer[index].to_i} #convert to integers
					@settings['tracks_to_rip'] -= answer # don't you just love ruby math? [1,2,3,4,5] - [3,4] = [1,2,5]
				elsif number == 2
					print "Type the numbers of the tracks you want to keep and separate them with a space: "
					answer = STDIN.gets.strip.split
					answer.each_index{|index| answer[index] = answer[index].to_i} #convert to integers
					remove = @settings['tracks_to_rip'] - answer # remove is inverted result -> which tracks you want to remove?
					@settings['tracks_to_rip'] -= remove # remove these
				else
					puts "#{number} is not a valid number! Please enter 1 or 2!\n"
				end
				if number == 1 || number == 2
					puts "Current selection of tracks : #{@settings['tracks_to_rip'].join(' ')}"
					succes = !get_answer("Do you want to make any changes? (y/n) : ", "yes", "n")
				end
			end
		end
	end
end	

Gui_CLI.new()
